
/*
 * $Id: EmbedContMdl.cc,v 1.4 1999/09/28 05:14:00 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include "Farm.h"
#include "ContBuildIter.h"
#include "ContentCfg.h"
#include "ContentMgr.h"

#include "ContentSym.h"

#include "EmbedContMdl.h"
#include "rndDistrStat.h"


static ObjFarm<ContBuildIter> TheIterators;


EmbedContMdl::EmbedContMdl(): theCfgSel(0), theEmbedCount(0),
	theChoiceSpace(1000), theEmbedCountMean(-1) {
	TheIterators.limit(1024); // magic, no good way to estimate
}


void EmbedContMdl::configure(const ContentSym *cfg) {
	Assert(!theEmbedCount);

	cfg->choiceSpace(theChoiceSpace);
	if ((theEmbedCount = cfg->embedCount()))
		theEmbedCountMean = RndDistrStat(theEmbedCount).mean();

	configureEmbedCfgs(cfg);

	if (theEmbedCount && !theCfgs.count()) {
		cerr << cfg->loc() 
			<< "the number of embedded objects per container is specified, "
			<< "but object content types are undefined"
			<< endl;
		exit(-2);
	}

	if (!theEmbedCount && theCfgs.count()) {
		cerr << cfg->loc() 
			<< "object content types are specified, "
			<< "but the number of embedded objects per container is not"
			<< endl;
		exit(-2);
	}
}

double EmbedContMdl::compLocalContrib(ContentCfg *cc) const {
	int idx = 0;
	if (find(cc->id(), idx))
		return theProbs[idx];
	return 0.0;
}

// average number of unique embedded objects per container
// given sliding factor k, average number of objects per container n,
// and choice space w
inline
double myF(int k, double n, int w) {
	const double q = pow(1 - 1./w, n*w/k);
	return k * (1 - q);
}

void EmbedContMdl::noteGlobalContrib(ContentCfg *cc, double cc_prob, double contrib) {
	int idx = 0;
	if (!find(cc->id(), idx))
		return;

	/* compude slide factor (a.k.a. "k"):
		u = f(k,n,w);   k <= w, f is ascending
		u,n,w are known
		find integer k that fits the best
	*/

	const double u = cc_prob / contrib;
	const double n = theEmbedCountMean * theProbs[idx];
	const int w = theChoiceSpace;

	int k;
	for (k = 1; k <= w && myF(k, n, w) <= u; ++k) {}

	// "p(k-1) <= u < p(k)": find closest end
	if (k == 1 || k == w)
		theSlides[idx] = k;
	else
	if (u - myF(k-1, n, w) <= myF(k, n, w) - u)
		theSlides[idx] = k-1;
	else
		theSlides[idx] = k;

	Assert(theSlides[idx] > 0);
}

ContBuildIter *EmbedContMdl::getIterator(Size contSize, WrBuf *buf) {
	ContBuildIter *i = TheIterators.get();
	i->configure(this, buf, contSize);
	return i;
}

void EmbedContMdl::putIterator(ContBuildIter *i) {
	TheIterators.put(i);
}

int EmbedContMdl::embedGoal(RndGen &rng) {
	if (!theEmbedCount)
		return 0;
	theEmbedCount->rndGen(&rng); // note: we do not reset rng back
	return (int)theEmbedCount->trial();
}

int EmbedContMdl::embedOid(RndGen &rng, int oid, int &embedType) {
	Assert(theEmbedCount);

	// first select the embeded type
	theCfgSel->rndGen(&rng);   // note: we do not reset rng back
	const int idx = (int)theCfgSel->trial();
	Assert(0 <= idx && idx < theCfgs.count());
	embedType = theCfgs[idx]->id();

	// compute embedded oid
	const int left = theSlides[idx] * oid;
	return rng(left, left + theChoiceSpace);
}

void EmbedContMdl::configureEmbedCfgs(const ContentSym *cfg) {
	Array<ContentSym*> syms;
	cfg->contains(syms, theCfgSel, theProbs);

	theCfgs.stretch(syms.count());
	for (int i = 0; i < syms.count(); ++i) {
		const ContentCfg *cfg = TheContentMgr.get(syms[i]);
		theCfgs.append(cfg);
	}

	theSlides.stretch(syms.count());
	theSlides.count(syms.count());
}

bool EmbedContMdl::find(int cfgId, int &idx) const {
	for (idx = 0; idx < theCfgs.count(); ++idx) {
		if (theCfgs[idx]->id() == cfgId)
			return true;
	}
	return false;
}
