
/*
 * $Id: udp2tcpd.cc,v 1.5 1999/09/24 21:48:17 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include <unistd.h>
#include <string.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <iostream.h>
#include <iomanip.h>

#include "Assert.h"
#include "Socket.h"
#include "Clock.h"
#include "Poll.h"
#include "NotifMsg.h"
#include "gadgets.h"

// maintains info about a host
class Client: public FileScanUser {
	public:
		Client(Socket aFD, const NetAddr &anAddr);
		virtual ~Client();

		void append(const void *buf, int size);

		virtual void noteWriteReady(int fd);
		
	protected:
		void selfDistruct();

	protected:
		Socket theSock;
		NetAddr theAddr; // unique host address
		FileScanReserv theReserv;
		char theBuf[1024];
		int theSize;
};

class TcpServer: public FileScanUser {
	public:
		TcpServer(int port);
		virtual ~TcpServer();

		virtual void noteReadReady(int fd);

	protected:
		Socket theSock;
		FileScanReserv theReserv;
};

class UdpServer: public FileScanUser {
	public:
		UdpServer(int port);
		virtual ~UdpServer();

		virtual void noteReadReady(int fd);

	protected:
		Socket theSock;
		FileScanReserv theReserv;
};


/* globals */

static Array<Client*> TheClients;
static FileScanner *TheScanner = 0;

static int TheLiveCltCnt = 0;

/* implementation */


/* Client */

Client::Client(Socket aFD, const NetAddr &anAddr): 
	theSock(aFD), theAddr(anAddr), theSize(0) {
	theSock.blocking(false);
	TheLiveCltCnt++;
}

Client::~Client() {
	if (theReserv)
		TheScanner->clearRes(theReserv);
	TheLiveCltCnt--;
}

void Client::append(const void *buf, int size) {
	if (size + theSize > (int)sizeof(theBuf)) {
		selfDistruct();
		return;
	}
	memcpy(theBuf + theSize, buf, size);
	theSize += size;
	if (!theReserv)
		theReserv = TheScanner->setFD(theSock.fd(), dirWrite, this);
}

void Client::noteWriteReady(int) {
	if (theSize) {
		const int sz = theSock.write(theBuf, theSize);
		if (sz < 0) {
			if (errno != EAGAIN)
				selfDistruct();
			return;
		}
		theSize -= sz;
		memmove(theBuf, theBuf+sz, theSize);
	}

	if (!theSize)
		TheScanner->clearRes(theReserv);
}

void Client::selfDistruct() {
	for (int i = 0; i < TheClients.count(); ++i)
		if (TheClients[i] == this) {
			TheClients[i] = 0;
			delete this;
			return;
		}
	Assert(0);
}


/* TcpServer */

TcpServer::TcpServer(int port) {
	Check(theSock.create());
	Check(theSock.blocking(false));
	theSock.reuseAddr(true);
	if (!theSock.bind(NetAddr(0, port)) || !theSock.listen()) {
		cerr << "cannot listen on TCP port " << port << ": " << error(errno) << endl;
		exit(-2);
	}

	theReserv = TheScanner->setFD(theSock.fd(), dirRead, this);
}

TcpServer::~TcpServer() {
	if (theReserv)
		TheScanner->clearRes(theReserv);
}

void TcpServer::noteReadReady(int) {
	sockaddr_in addr;
	socklen_t addr_len = sizeof(addr);
	if (Socket s = theSock.accept((struct sockaddr*)&addr, &addr_len)) {
		Client *c = new Client(s, NetAddr(
			((sockaddr_in &)addr).sin_addr,
			((sockaddr_in &)addr).sin_port));

		for (int i = 0; i < TheClients.count(); ++i)
			if (!TheClients[i]) {
				TheClients[i] = c;
				return;
			}
		TheClients.append(c);
	}
}


/* UdpServer */

UdpServer::UdpServer(int port) {
	Check(theSock.create(AF_INET, SOCK_DGRAM, 0));
	theSock.reuseAddr(true);
	if (!theSock.bind(NetAddr(0, port))) {
		cerr << "cannot listen on UDP port " << port << ": " << error(errno) << endl;
		exit(-2);
	}
	Check(theSock.blocking(false));

	theReserv = TheScanner->setFD(theSock.fd(), dirRead, this);
}

UdpServer::~UdpServer() {
	if (theReserv)
		TheScanner->clearRes(theReserv);
}

void UdpServer::noteReadReady(int) {
	StatusNotifMsg msg;
	struct sockaddr from;
	socklen_t fromlen = sizeof(from);
	if (recvfrom(theSock.fd(), (char*)&msg, sizeof(msg), 0, &from, &fromlen) == sizeof(msg)) {
		msg.ntoh();
		StatusFwdMsg fwd(msg, TheClock, NetAddr((sockaddr_in &)from));
		fwd.theCopyCnt = TheLiveCltCnt;
		fwd.hton();
		for (int i = 0; i < TheClients.count(); ++i)
			if (TheClients[i])
				TheClients[i]->append(&fwd, sizeof(fwd));
	}
}


int main(int argc, char *argv[]) {

	// note: we assume that TCP and UDP port namespaces are separate
	if (argc > 2 || (argc == 2 && !xatoi(argv[1]))) {
		cerr << "usage: " << argv[0] << " [udp_and_tcp_port]" << endl;
		return -1;
	}

	const int port = xatoi(argv[1], 18256);

	cout << "Starting UDP to TCP dispatcher on port " << port << endl;

	// detach ourselves
	const pid_t pid = fork();

	if (pid > 0) // parent
		return 0;

    if (pid < 0 || setsid() < 0) { // error
        cerr << "problems detaching the server process: " << error(errno) << endl;
		return -1;
	}

	TheScanner = new Poll();
	TheScanner->configure(FD_SETSIZE);

	TcpServer *tcpSrv = new TcpServer(port);
	UdpServer *udpSrv = new UdpServer(port);

	signal(SIGPIPE, SIG_IGN);
	signal(SIGHUP, SIG_IGN);

	while (!TheScanner->idle()) {
		Clock::Update();
		TheScanner->scan(0);
	}

	delete udpSrv;
	delete tcpSrv;
	delete TheScanner;

	return 0;
}
