
/*
 * $Id: Rnd.h,v 1.6 1999/09/28 04:01:04 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#ifndef XSTD_RND_CLASS_H
#define XSTD_RND_CLASS_H

#include <math.h>

class ostream;

// simple random generator
// every generator gets its own seed and state
// no virtual stuff as we do not need more than one type of generators
class RndGen {
	public:
		typedef long State;
		static State TheDefSeed;   // default seed

	public:
		RndGen(State aSeed = -1);

		// uniform[0,1)
		double trial() { return ltrial()/2147483648.0; }
		double operator ()() { return trial(); }

		// uniform[0,2147483648)
		long ltrial();

		// equiprob[lo,hi); "hi" is excluded inless hi==lo
		double operator ()(double lo, double hi) { return lo + (hi-lo)*trial(); }
		int operator ()(int lo, int hi) { return hi != lo ? lo + ltrial()%(hi-lo) : lo; }

		// Bernoulli event with a given probability
		bool event(double prob) { return trial() < prob; }

		// change seed
		void seed(State seed);
		void reset() { seed(theSeed); }

	protected:
		State theSeed;  // initial seed
		State theState; // current state
};

// abstract distribution
// allows for sharing of one random generator among several distributions
class RndDistr
{
	public:
		RndDistr(RndGen *aGen): theGen(aGen) {}
		virtual ~RndDistr() {} // may be leaking; we do not delete theGen

		virtual const char *pdfName() const = 0; // type of pdf (e.g., exp)
		RndGen *rndGen() { return theGen; }
		virtual void rndGen(RndGen *aGen) { theGen = aGen; }

		// returns next pseudo random number
		double operator ()() { return trial(); }
		virtual double trial() = 0;

		virtual double mean() const = 0; // mean
		virtual double sdev() const = 0;  // standard deviation
		double var() const { return sdev()*sdev(); }  // variance

		virtual ostream &print(ostream &os) const;

	protected:
		RndGen *theGen;
};

inline
ostream &operator <<(ostream &os, const RndDistr &d) { return d.print(os); }

#endif
